""" DIRAC FileCatalog plug-in class to manage dynamic datasets defined by a metadata query
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import hashlib
import os

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.List import stringListToString


class DatasetManager(object):

    _tables = dict()
    _tables["FC_MetaDatasets"] = {
        "Fields": {
            "DatasetID": "INT AUTO_INCREMENT",
            "DatasetName": "VARCHAR(128) CHARACTER SET latin1 COLLATE latin1_bin NOT NULL",
            "MetaQuery": "VARCHAR(512)",
            "DirID": "INT NOT NULL DEFAULT 0",
            "TotalSize": "BIGINT UNSIGNED NOT NULL",
            "NumberOfFiles": "INT NOT NULL",
            "UID": "SMALLINT UNSIGNED NOT NULL",
            "GID": "TINYINT UNSIGNED NOT NULL",
            "Status": "SMALLINT UNSIGNED NOT NULL",
            "CreationDate": "DATETIME",
            "ModificationDate": "DATETIME",
            "DatasetHash": "CHAR(36) NOT NULL",
            "Mode": "SMALLINT UNSIGNED NOT NULL DEFAULT 509",
        },
        "UniqueIndexes": {"DatasetName_DirID": ["DatasetName", "DirID"]},
        "PrimaryKey": "DatasetID",
    }
    _tables["FC_MetaDatasetFiles"] = {
        "Fields": {
            "DatasetID": "INT NOT NULL",
            "FileID": "INT NOT NULL",
        },
        "UniqueIndexes": {"DatasetID_FileID": ["DatasetID", "FileID"]},
    }
    _tables["FC_DatasetAnnotations"] = {
        "Fields": {"DatasetID": "INT NOT NULL", "Annotation": "VARCHAR(512)"},
        "PrimaryKey": "DatasetID",
    }

    def __init__(self, database=None):
        self.db = None
        if database is not None:
            self.setDatabase(database)

    def setDatabase(self, database):
        self.db = database

        # Have to check the existing table should be dropped when
        # existing tables will not be recreated
        result = self.db._query("SHOW TABLES")
        if not result["OK"]:
            return result
        tableList = [x[0] for x in result["Value"]]
        tablesToCreate = {}
        for table in self._tables:
            if table not in tableList:
                tablesToCreate[table] = self._tables[table]

        result = self.db._createTables(tablesToCreate)
        if not result["OK"]:
            gLogger.error("Failed to create tables", str(self._tables.keys()))
        elif result["Value"]:
            gLogger.info("Tables created: %s" % ",".join(result["Value"]))
        return result

    def _getConnection(self, connection=False):
        if connection:
            return connection
        res = self.db._getConnection()
        if res["OK"]:
            return res["Value"]
        return connection

    def addDataset(self, datasets, credDict):
        """
        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """

        result = self.db.ugManager.getUserAndGroupID(credDict)
        if not result["OK"]:
            return result
        uid, gid = result["Value"]

        failed = dict()
        successful = dict()

        for datasetName, metaQuery in datasets.items():
            result = self.__addDataset(datasetName, metaQuery, credDict, uid, gid)
            if result["OK"]:
                successful[datasetName] = True
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __addDataset(self, datasetName, metaQuery, credDict, uid, gid):

        result = self.__getMetaQueryParameters(metaQuery, credDict)
        if not result["OK"]:
            return result
        totalSize = result["Value"]["TotalSize"]
        datasetHash = result["Value"]["DatasetHash"]
        numberOfFiles = result["Value"]["NumberOfFiles"]

        result = self.db.fileManager._getStatusInt("Dynamic")
        if not result["OK"]:
            return result
        intStatus = result["Value"]

        dsDir = os.path.dirname(datasetName)
        dsName = os.path.basename(datasetName)
        if not dsDir:
            return S_ERROR("Dataset name should be specified with full path")
        result = self.db.dtree.existsDir(dsDir)
        if not result["OK"]:
            return result
        if result["Value"]["Exists"]:
            dirID = result["Value"]["DirID"]
        else:
            result = self.db.dtree.makeDirectories(dsDir)
            if not result["OK"]:
                return result
            dirID = result["Value"]

        # Add the new dataset entry now
        inDict = {
            "DatasetName": dsName,
            "MetaQuery": str(metaQuery),
            "DirID": dirID,
            "TotalSize": totalSize,
            "NumberOfFiles": numberOfFiles,
            "UID": uid,
            "GID": gid,
            "CreationDate": "UTC_TIMESTAMP()",
            "ModificationDate": "UTC_TIMESTAMP()",
            "DatasetHash": datasetHash,
            "Status": intStatus,
        }
        result = self.db.insertFields("FC_MetaDatasets", inDict=inDict)
        if not result["OK"]:
            if "Duplicate" in result["Message"]:
                return S_ERROR("Dataset %s already exists" % datasetName)
            else:
                return result
        datasetID = result["lastRowId"]
        return S_OK(datasetID)

    def _getDatasetDirectories(self, datasets):
        dirDict = {}
        for path in datasets:
            dsDir = os.path.dirname(path)
            dsFile = os.path.basename(path)
            dirDict.setdefault(dsDir, [])
            dirDict[dsDir].append(dsFile)
        return dirDict

    def _findDatasets(self, datasets, connection=False):

        connection = self._getConnection(connection)
        fullNames = [name for name in datasets if name.startswith("/")]
        shortNames = [name for name in datasets if not name.startswith("/")]

        resultDict = {"Successful": {}, "Failed": {}}
        if fullNames:
            result = self.__findFullPathDatasets(fullNames, connection)
            if not result["OK"]:
                return result
            resultDict = result["Value"]

        if shortNames:
            result = self.__findNoPathDatasets(shortNames, connection)
            if not result["OK"]:
                return result
            resultDict["Successful"].update(result["Value"]["Successful"])
            resultDict["Failed"].update(result["Value"]["Failed"])

        return S_OK(resultDict)

    def __findFullPathDatasets(self, datasets, connection):

        dirDict = self._getDatasetDirectories(datasets)
        failed = {}
        successful = {}
        result = self.db.dtree.findDirs(list(dirDict))
        if not result["OK"]:
            return result
        directoryIDs = result["Value"]

        directoryPaths = {}
        for dirPath in dirDict:
            if dirPath not in directoryIDs:
                for dsName in dirDict[dirPath]:
                    dname = "%s/%s" % (dirPath, dsName)
                    dname = dname.replace("//", "/")
                    failed[dname] = "No such dataset or directory"
            else:
                directoryPaths[directoryIDs[dirPath]] = dirPath

        wheres = []
        for dirPath in directoryIDs:
            dsNames = dirDict[dirPath]
            dirID = directoryIDs[dirPath]
            wheres.append("( DirID=%d AND DatasetName IN (%s) )" % (dirID, stringListToString(dsNames)))

        req = "SELECT DatasetName,DirID,DatasetID FROM FC_MetaDatasets WHERE %s" % " OR ".join(wheres)
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result
        for dsName, dirID, dsID in result["Value"]:
            dname = "%s/%s" % (directoryPaths[dirID], dsName)
            dname = dname.replace("//", "/")
            successful[dname] = {"DirID": dirID, "DatasetID": dsID}

        for dataset in datasets:
            if dataset not in successful:
                failed[dataset] = "No such dataset"

        return S_OK({"Successful": successful, "Failed": failed})

    def __findNoPathDatasets(self, nodirDatasets, connection):

        failed = {}
        successful = {}
        dsIDs = {}
        req = "SELECT COUNT(DatasetName),DatasetName,DatasetID FROM FC_MetaDatasets WHERE DatasetName in "
        req += "( %s ) GROUP BY DatasetName,DatasetID" % stringListToString(nodirDatasets)
        result = self.db._query(req, connection)
        if not result["OK"]:
            return result
        for dsCount, dsName, dsID in result["Value"]:
            if dsCount > 1:
                failed[dsName] = "Ambiguous dataset name"
            else:
                dsIDs[dsName] = str(dsID)

        if dsIDs:
            req = "SELECT DatasetName,DatasetID,DirID FROM FC_MetaDatasets WHERE DatasetID in (%s)" % ",".join(
                dsIDs.values()
            )
            result = self.db._query(req, connection)
            if not result["OK"]:
                return result
            for dsName, dsID, dirID in result["Value"]:
                successful[dsName] = {"DirID": dirID, "DatasetID": dsID}

        for name in nodirDatasets:
            if name not in failed and name not in successful:
                failed[name] = "Dataset not found"

        return S_OK({"Successful": successful, "Failed": failed})

    def addDatasetAnnotation(self, datasets, credDict):
        """Add annotation to the given dataset"""
        connection = self._getConnection()
        successful = {}
        result = self._findDatasets(list(datasets), connection)
        if not result["OK"]:
            return result
        failed = result["Value"]["Failed"]
        datasetDict = result["Value"]["Successful"]
        for dataset, annotation in datasets.items():
            if dataset in datasetDict:
                req = "REPLACE FC_DatasetAnnotations (Annotation,DatasetID) VALUE ('%s',%d)" % (
                    annotation,
                    datasetDict[dataset]["DatasetID"],
                )
                result = self.db._update(req, connection)
                if not result["OK"]:
                    failed[dataset] = "Failed to add annotation"
                else:
                    successful[dataset] = True

        return S_OK({"Successful": successful, "Failed": failed})

    def getDatasetAnnotation(self, datasets, credDict={}):
        """Get annotations for the given datasets"""
        successful = {}
        failed = {}
        if datasets:
            result = self._findDatasets(datasets)
            if not result["OK"]:
                return result
            dsDict = {}
            for dataset in result["Value"]["Successful"]:
                dsDict[result["Value"]["Successful"][dataset]["DatasetID"]] = dataset
            for dataset in result["Value"]["Failed"]:
                failed[dataset] = "Dataset not found"

            idString = ",".join([str(x) for x in dsDict])
            req = "SELECT DatasetID, Annotation FROM FC_DatasetAnnotations WHERE DatasetID in (%s)" % idString
        else:
            req = "SELECT DatasetID, Annotation FROM FC_DatasetAnnotations"
        result = self.db._query(req)
        if not result["OK"]:
            return result

        for dsID, annotation in result["Value"]:
            successful[dsDict[dsID]] = annotation

        if datasets:
            for dataset in datasets:
                if dataset not in successful and dataset not in failed:
                    successful[dataset] = "-"

        return S_OK({"Successful": successful, "Failed": failed})

    def __getMetaQueryParameters(self, metaQuery, credDict):
        """Get parameters ( hash, total size, number of files ) for the given metaquery"""
        findMetaQuery = dict(metaQuery)

        path = "/"
        if "Path" in findMetaQuery:
            path = findMetaQuery["Path"]
            findMetaQuery.pop("Path")

        result = self.db.fmeta.findFilesByMetadata(findMetaQuery, path, credDict)
        if not result["OK"]:
            return S_ERROR("Failed to apply the metaQuery")

        idLfnDict = result["Value"]
        lfnIDList = list(idLfnDict)
        lfnList = sorted(idLfnDict.values())
        myMd5 = hashlib.md5()
        myMd5.update(str(lfnList).encode())
        datasetHash = myMd5.hexdigest().upper()
        numberOfFiles = len(lfnList)
        result = self.db.fileManager.getFileSize(lfnList)
        totalSize = 0
        if result["OK"]:
            totalSize = result["TotalSize"]

        result = S_OK(
            {
                "DatasetHash": datasetHash,
                "NumberOfFiles": numberOfFiles,
                "TotalSize": totalSize,
                "LFNList": lfnList,
                "LFNIDList": lfnIDList,
            }
        )
        return result

    def removeDataset(self, datasets, credDict):
        """Remove the requested datasets

        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """
        failed = dict()
        successful = dict()
        for datasetName in datasets:
            result = self.__removeDataset(datasetName, credDict)
            if result["OK"]:
                successful[datasetName] = True
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __removeDataset(self, datasetName, credDict):
        """Remove existing dataset"""

        req = "SELECT DatasetID FROM FC_MetaDatasets WHERE DatasetName='%s'" % datasetName
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            # No requested dataset
            return S_OK("Dataset %s does not exist" % datasetName)
        datasetID = result["Value"][0][0]

        for table in ["FC_MetaDatasetFiles", "FC_MetaDatasets", "FC_DatasetAnnotations"]:
            req = "DELETE FROM %s WHERE DatasetID=%s" % (table, datasetID)
            result = self.db._update(req)

        return result

    def checkDataset(self, datasets, credDict):
        """Check that the dataset parameters correspond to the actual state

        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """
        failed = dict()
        successful = dict()
        for datasetName in datasets.keys():
            result = self.__checkDataset(datasetName, credDict)
            if result["OK"]:
                successful[datasetName] = result["Value"]
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __checkDataset(self, datasetName, credDict):
        """Check that the dataset parameters correspond to the actual state"""
        req = "SELECT MetaQuery,DatasetHash,TotalSize,NumberOfFiles FROM FC_MetaDatasets"
        req += " WHERE DatasetName='%s'" % datasetName
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Unknown MetaDataset %s" % datasetName)

        row = result["Value"][0]
        metaQuery = eval(row[0])
        datasetHashOld = row[1]
        totalSizeOld = int(row[2])
        numberOfFilesOld = int(row[3])

        result = self.__getMetaQueryParameters(metaQuery, credDict)
        if not result["OK"]:
            return result
        totalSize = result["Value"]["TotalSize"]
        datasetHash = result["Value"]["DatasetHash"]
        numberOfFiles = result["Value"]["NumberOfFiles"]

        changeDict = {}
        if totalSize != totalSizeOld:
            changeDict["TotalSize"] = (totalSizeOld, totalSize)
        if datasetHash != datasetHashOld:
            changeDict["DatasetHash"] = (datasetHashOld, datasetHash)
        if numberOfFiles != numberOfFilesOld:
            changeDict["NumberOfFiles"] = (numberOfFilesOld, numberOfFiles)

        result = S_OK(changeDict)
        return result

    def updateDataset(self, datasets, credDict):
        """
        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """

        failed = dict()
        successful = dict()

        for datasetName in datasets.keys():
            result = self.__updateDataset(datasetName, credDict)
            if result["OK"]:
                successful[datasetName] = result["Value"]
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __updateDataset(self, datasetName, credDict):
        """Update the dataset parameters"""

        changeDict = {}
        result = self.__checkDataset(datasetName, credDict)
        if not result["OK"]:
            return result
        if not result["Value"]:
            # The dataset is not changed
            return S_OK()
        else:
            changeDict = result["Value"]

        req = "UPDATE FC_MetaDatasets SET "
        for field in changeDict:
            req += "%s='%s', " % (field, str(changeDict[field][1]))
        req += "ModificationDate=UTC_TIMESTAMP() "
        req += "WHERE DatasetName='%s'" % datasetName
        result = self.db._update(req)
        return result

    def getDatasets(self, datasets, credDict):
        """Get dataset definitions

        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """
        failed = dict()
        successful = dict()
        for datasetName in datasets:
            result = self.__getDatasets(datasetName)
            if result["OK"]:
                successful[datasetName] = result["Value"]
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __getDatasets(self, datasetName):
        """Get information about existing datasets"""

        parameterList = [
            "DatasetID",
            "MetaQuery",
            "DirID",
            "TotalSize",
            "NumberOfFiles",
            "UID",
            "GID",
            "Status",
            "CreationDate",
            "ModificationDate",
            "DatasetHash",
            "Mode",
            "DatasetName",
        ]
        parameterString = ",".join(parameterList)

        req = "SELECT %s FROM FC_MetaDatasets" % parameterString
        dsName = os.path.basename(datasetName)
        if "*" in dsName:
            dName = dsName.replace("*", "%")
            req += " WHERE DatasetName LIKE '%s'" % dName
        elif dsName:
            req += " WHERE DatasetName='%s'" % dsName

        result = self.db._query(req)
        if not result["OK"]:
            return result

        resultDict = {}
        for row in result["Value"]:
            dName = row[12]
            resultDict[dName] = self.__getDatasetDict(row)

        return S_OK(resultDict)

    def getDatasetsInDirectory(self, dirID, verbose=False, connection=False):
        """Get datasets in the given directory"""
        datasets = {}
        parameterList = [
            "DatasetName",
            "DatasetID",
            "MetaQuery",
            "DirID",
            "TotalSize",
            "NumberOfFiles",
            "UID",
            "GID",
            "Status",
            "CreationDate",
            "ModificationDate",
            "DatasetHash",
            "Mode",
        ]
        parameterString = ",".join(parameterList)

        req = "SELECT %s FROM FC_MetaDatasets WHERE DirID=%s" % (parameterString, str(dirID))
        result = self.db._query(req)
        if not result["OK"]:
            return result
        userDict = {}
        groupDict = {}
        for row in result["Value"]:
            dsName = row[0]
            datasets[dsName] = {}
            dsDict = {}
            for i in range(1, len(row), 1):
                dsDict[parameterList[i]] = row[i]
                if parameterList[i] == "UID":
                    uid = row[i]
                    if uid in userDict:
                        owner = userDict[uid]
                    else:
                        owner = "unknown"
                        result = self.db.ugManager.getUserName(uid)
                        if result["OK"]:
                            owner = result["Value"]
                        userDict[uid] = owner
                    dsDict[dsName]["Owner"] = owner
                if parameterList[i] == "GID":
                    gid = row[i]
                    if gid in groupDict:
                        group = groupDict[gid]
                    else:
                        group = "unknown"
                        result = self.db.ugManager.getGroupName(gid)
                        if result["OK"]:
                            group = result["Value"]
                        groupDict[gid] = group
                    dsDict[dsName]["OwnerGroup"] = group
            datasets[dsName]["Metadata"] = dsDict

        if verbose and datasets:
            result = self.getDatasetAnnotation(list(datasets))
            if result["OK"]:
                for dataset in result["Value"]["Successful"]:
                    datasets[dataset]["Annotation"] = result["Value"]["Successful"][dataset]
                for dataset in result["Value"]["Failed"]:
                    datasets[dataset]["Annotation"] = result["Value"]["Failed"][dataset]

        return S_OK(datasets)

    def __getDatasetDict(self, row):

        resultDict = {}
        resultDict["DatasetID"] = int(row[0])
        resultDict["MetaQuery"] = eval(row[1])
        resultDict["DirID"] = int(row[2])
        resultDict["TotalSize"] = int(row[3])
        resultDict["NumberOfFiles"] = int(row[4])
        uid = int(row[5])
        gid = int(row[6])
        result = self.db.ugManager.getUserName(uid)
        if result["OK"]:
            resultDict["Owner"] = result["Value"]
        else:
            resultDict["Owner"] = "Unknown"
        result = self.db.ugManager.getGroupName(gid)
        if result["OK"]:
            resultDict["OwnerGroup"] = result["Value"]
        else:
            resultDict["OwnerGroup"] = "Unknown"
        intStatus = int(row[7])
        result = self.db.fileManager._getIntStatus(intStatus)
        if result["OK"]:
            resultDict["Status"] = result["Value"]
        else:
            resultDict["Status"] = "Unknown"
        resultDict["CreationDate"] = row[8]
        resultDict["ModificationDate"] = row[9]
        resultDict["DatasetHash"] = row[10]
        resultDict["Mode"] = row[11]

        return resultDict

    def getDatasetParameters(self, datasets, credDict):
        """Get dataset definitions

        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """
        failed = dict()
        successful = dict()
        for datasetName in datasets:
            result = self.__getDatasetParameters(datasetName, credDict)
            if result["OK"]:
                successful[datasetName] = result["Value"]
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __getDatasetParameters(self, datasetName, credDict):
        """Get the currently stored dataset parameters"""
        result = self._findDatasets([datasetName])
        if not result["OK"]:
            return result
        if not result["Value"]["Successful"]:
            return S_ERROR(result["Value"]["Failed"][datasetName])
        dirID = result["Value"]["Successful"][datasetName]["DirID"]
        dsName = os.path.basename(datasetName)

        parameterList = [
            "DatasetID",
            "MetaQuery",
            "DirID",
            "TotalSize",
            "NumberOfFiles",
            "UID",
            "GID",
            "Status",
            "CreationDate",
            "ModificationDate",
            "DatasetHash",
            "Mode",
        ]
        parameterString = ",".join(parameterList)

        req = "SELECT %s FROM FC_MetaDatasets WHERE DatasetName='%s' AND DirID=%d" % (parameterString, dsName, dirID)
        result = self.db._query(req)
        if not result["OK"]:
            return result

        row = result["Value"][0]
        resultDict = self.__getDatasetDict(row)

        return S_OK(resultDict)

    def setDatasetStatus(self, datasetName, status):
        """Set the given dataset status"""
        result = self.db.fileManager._getStatusInt(status)
        if not result["OK"]:
            return result
        intStatus = result["Value"]
        req = "UPDATE FC_MetaDatasets SET Status=%d, ModificationDate=UTC_TIMESTAMP() " % intStatus
        req += "WHERE DatasetName='%s'" % datasetName
        result = self.db._update(req)
        return result

    def getDatasetStatus(self, datasetName, credDict):
        """Get status of the given dataset"""

        result = self.__getDatasetParameters(datasetName, credDict)
        if not result["OK"]:
            return result
        status = result["Value"]["Status"]
        return S_OK(status)

    def __getDynamicDatasetFiles(self, datasetID, credDict):
        """Get dataset lfns from a dynamic meta query"""
        req = "SELECT MetaQuery FROM FC_MetaDatasets WHERE DatasetID=%d" % datasetID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Unknown MetaDataset ID %d" % datasetID)

        metaQuery = eval(result["Value"][0][0])
        result = self.__getMetaQueryParameters(metaQuery, credDict)
        if not result["OK"]:
            return result

        lfnList = result["Value"]["LFNList"]
        finalResult = S_OK(lfnList)
        finalResult["FileIDList"] = result["Value"]["LFNIDList"]
        return finalResult

    def __getFrozenDatasetFiles(self, datasetID, credDict):
        """Get dataset lfns from a frozen snapshot"""

        req = "SELECT FileID FROM FC_MetaDatasetFiles WHERE DatasetID=%d" % datasetID
        result = self.db._query(req)
        if not result["OK"]:
            return result

        fileIDList = [row[0] for row in result["Value"]]
        result = self.db.fileManager._getFileLFNs(fileIDList)
        if not result["OK"]:
            return result

        lfnDict = result["Value"]["Successful"]
        fileIDList = list(lfnDict)
        result = S_OK([lfnDict[i] for i in fileIDList])
        result["FileIDList"] = fileIDList
        return result

    def getDatasetFiles(self, datasets, credDict):
        """Get dataset file contents

        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """
        failed = dict()
        successful = dict()
        for datasetName in datasets:
            result = self.__getDatasetFiles(datasetName, credDict)
            if result["OK"]:
                successful[datasetName] = result["Value"]
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __getDatasetFiles(self, datasetName, credDict):
        """Get dataset files"""
        result = self.__getDatasetParameters(datasetName, credDict)
        if not result["OK"]:
            return result
        status = result["Value"]["Status"]
        datasetID = result["Value"]["DatasetID"]
        if status in ["Frozen", "Static"]:
            return self.__getFrozenDatasetFiles(datasetID, credDict)
        else:
            return self.__getDynamicDatasetFiles(datasetID, credDict)

    def freezeDataset(self, datasets, credDict):
        """Freeze the contents of datasets

        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """
        failed = dict()
        successful = dict()
        for datasetName in datasets:
            result = self.__freezeDataset(datasetName, credDict)
            if result["OK"]:
                successful[datasetName] = True
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __freezeDataset(self, datasetName, credDict):
        """Freeze the contents of the dataset"""
        result = self.__getDatasetParameters(datasetName, credDict)
        if not result["OK"]:
            return result
        status = result["Value"]["Status"]
        if status == "Frozen":
            return S_OK()

        datasetID = result["Value"]["DatasetID"]
        req = "DELETE FROM FC_MetaDatasetFiles WHERE DatasetID=%d" % datasetID
        result = self.db._update(req)

        result = self.__getDynamicDatasetFiles(datasetID, credDict)

        if not result["OK"]:
            return result
        fileIDList = result["FileIDList"]
        valueList = []
        for fileID in fileIDList:
            valueList.append("(%d,%d)" % (datasetID, fileID))
        valueString = ",".join(valueList)
        req = "INSERT INTO FC_MetaDatasetFiles (DatasetID,FileID) VALUES %s" % valueString
        result = self.db._update(req)
        if not result["OK"]:
            return result

        result = self.setDatasetStatus(datasetName, "Frozen")
        return result

    def releaseDataset(self, datasets, credDict):
        """Unfreeze datasets

        :param dict datasets: dictionary describing dataset definitions
        :param credDict:  dictionary of the caller credentials
        :return: S_OK/S_ERROR bulk return structure
        """
        failed = dict()
        successful = dict()
        for datasetName in datasets:
            result = self.__releaseDataset(datasetName, credDict)
            if result["OK"]:
                successful[datasetName] = True
            else:
                failed[datasetName] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __releaseDataset(self, datasetName, credDict):
        """return the dataset to a dynamic state"""
        result = self.__getDatasetParameters(datasetName, credDict)
        if not result["OK"]:
            return result
        status = result["Value"]["Status"]
        if status == "Dynamic":
            return S_OK()

        datasetID = result["Value"]["DatasetID"]
        req = "DELETE FROM FC_MetaDatasetFiles WHERE DatasetID=%d" % datasetID
        result = self.db._update(req)

        result = self.setDatasetStatus(datasetName, "Dynamic")
        return result



"""Service interface is the service which provide config for client and synchronize Master/Slave servers"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import time
import re
import zipfile
import zlib

import DIRAC
from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData, ConfigurationData
from DIRAC.ConfigurationSystem.private.Refresher import gRefresher
from DIRAC.Core.Base.Client import Client
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.FrameworkSystem.Client.Logger import gLogger


class ServiceInterfaceBase(object):
    """Service interface is the service which provide config for client and synchronize Master/Slave servers"""

    def __init__(self, sURL):
        self.sURL = sURL
        gLogger.info("Initializing Configuration Service", "URL is %s" % sURL)
        self.__modificationsIgnoreMask = ["/DIRAC/Configuration/Servers", "/DIRAC/Configuration/Version"]
        gConfigurationData.setAsService()
        if not gConfigurationData.isMaster():
            gLogger.info("Starting configuration service as slave")
            gRefresher.autoRefreshAndPublish(self.sURL)
        else:
            gLogger.info("Starting configuration service as master")
            gRefresher.disable()
            self.__loadConfigurationData()
            self.dAliveSlaveServers = {}
            self._launchCheckSlaves()

    def isMaster(self):
        return gConfigurationData.isMaster()

    def _launchCheckSlaves(self):
        raise NotImplementedError("Should be implemented by the children class")

    def __loadConfigurationData(self):
        """
        As the name said, it just load the configuration
        """

        mkDir(os.path.join(DIRAC.rootPath, "etc", "csbackup"))
        gConfigurationData.loadConfigurationData()
        if gConfigurationData.isMaster():
            bBuiltNewConfiguration = False
            if not gConfigurationData.getName():
                DIRAC.abort(10, "Missing name for the configuration to be exported!")
            gConfigurationData.exportName()
            sVersion = gConfigurationData.getVersion()
            if sVersion == "0":
                gLogger.info("There's no version. Generating a new one")
                gConfigurationData.generateNewVersion()
                bBuiltNewConfiguration = True

            if self.sURL not in gConfigurationData.getServers():
                gConfigurationData.setServers(self.sURL)
                bBuiltNewConfiguration = True

            gConfigurationData.setMasterServer(self.sURL)

            if bBuiltNewConfiguration:
                gConfigurationData.writeRemoteConfigurationToDisk()

    def __generateNewVersion(self):
        """
        After changing configuration, we use this method to save them
        """
        if gConfigurationData.isMaster():
            gConfigurationData.generateNewVersion()
            gConfigurationData.writeRemoteConfigurationToDisk()

    def publishSlaveServer(self, sSlaveURL):
        """
        Called by the slave server via service, it register a new slave server

        :param sSlaveURL: url of slave server
        """

        if not gConfigurationData.isMaster():
            return S_ERROR("Configuration modification is not allowed in this server")
        gLogger.info("Pinging slave %s" % sSlaveURL)
        rpcClient = ConfigurationClient(url=sSlaveURL, timeout=10, useCertificates=True)
        retVal = rpcClient.ping()
        if not retVal["OK"]:
            gLogger.info("Slave %s didn't reply" % sSlaveURL)
            return
        if retVal["Value"]["name"] != "Configuration/Server":
            gLogger.info("Slave %s is not a CS serveR" % sSlaveURL)
            return
        bNewSlave = False
        if sSlaveURL not in self.dAliveSlaveServers:
            bNewSlave = True
            gLogger.info("New slave registered", sSlaveURL)
        self.dAliveSlaveServers[sSlaveURL] = time.time()
        if bNewSlave:
            gConfigurationData.setServers(", ".join(self.dAliveSlaveServers))
            self.__generateNewVersion()

    def _checkSlavesStatus(self, forceWriteConfiguration=False):
        """
        Check if Slaves server are still availlable

        :param forceWriteConfiguration: (default False) Force rewriting configuration after checking slaves
        """

        gLogger.info("Checking status of slave servers")
        iGraceTime = gConfigurationData.getSlavesGraceTime()
        bModifiedSlaveServers = False
        for sSlaveURL in list(self.dAliveSlaveServers):
            if time.time() - self.dAliveSlaveServers[sSlaveURL] > iGraceTime:
                gLogger.warn("Found dead slave", sSlaveURL)
                del self.dAliveSlaveServers[sSlaveURL]
                bModifiedSlaveServers = True
        if bModifiedSlaveServers or forceWriteConfiguration:
            gConfigurationData.setServers(", ".join(self.dAliveSlaveServers))
            self.__generateNewVersion()

    @staticmethod
    def _forceServiceUpdate(url, fromMaster):
        """
        Force updating configuration on a given service
        This should be called by _updateServiceConfiguration

        :param str url: service URL
        :param bool fromMaster: flag to force updating from the master CS
        :return: S_OK/S_ERROR
        """
        gLogger.info("Updating service configuration on", url)

        result = Client(url=url).refreshConfiguration(fromMaster)
        result["URL"] = url
        return result

    def _updateServiceConfiguration(self, urlSet, fromMaster=False):
        """
        Update configuration in a set of service in parallel

        :param set urlSet: a set of service URLs
        :param fromMaster: flag to force updating from the master CS
        :return: Nothing
        """
        raise NotImplementedError("Should be implemented by the children class")

    def forceSlavesUpdate(self):
        """
        Force updating configuration on all the slave configuration servers

        :return: Nothing
        """
        gLogger.info("Updating configuration on slave servers")
        iGraceTime = gConfigurationData.getSlavesGraceTime()
        urlSet = set()
        for slaveURL in self.dAliveSlaveServers:
            if time.time() - self.dAliveSlaveServers[slaveURL] <= iGraceTime:
                urlSet.add(slaveURL)
        self._updateServiceConfiguration(urlSet, fromMaster=True)

    def forceGlobalUpdate(self):
        """
        Force updating configuration of all the registered services

        :returns: S_OK (needed for DISET return call)
        """
        gLogger.info("Updating services configuration")
        # Get URLs of all the services except for Configuration services
        cfg = gConfigurationData.remoteCFG.getAsDict()["Systems"]
        urlSet = set()
        for system_ in cfg:
            for instance in cfg[system_]:
                for url in cfg[system_][instance]["URLs"]:
                    urlSet = urlSet.union(
                        set(
                            [
                                u.strip()
                                for u in cfg[system_][instance]["URLs"][url].split(",")
                                if "Configuration/Server" not in u
                            ]
                        )
                    )
        self._updateServiceConfiguration(urlSet)
        return S_OK()

    def updateConfiguration(self, sBuffer, committer="", updateVersionOption=False):
        """
        Update the master configuration with the newly received changes

        :param str sBuffer: newly received configuration data
        :param str committer: the user name of the committer
        :param bool updateVersionOption: flag to update the current configuration version
        :return: S_OK/S_ERROR of the write-to-disk of the new configuration
        """
        if not gConfigurationData.isMaster():
            return S_ERROR("Configuration modification is not allowed in this server")
        # Load the data in a ConfigurationData object
        oRemoteConfData = ConfigurationData(False)
        oRemoteConfData.loadRemoteCFGFromCompressedMem(sBuffer)
        if updateVersionOption:
            oRemoteConfData.setVersion(gConfigurationData.getVersion())
        # Test that remote and new versions are the same
        sRemoteVersion = oRemoteConfData.getVersion()
        sLocalVersion = gConfigurationData.getVersion()
        gLogger.info("Checking versions\n", "remote: %s\nlocal:  %s" % (sRemoteVersion, sLocalVersion))
        if sRemoteVersion != sLocalVersion:
            if not gConfigurationData.mergingEnabled():
                return S_ERROR(
                    "Local and remote versions differ (%s vs %s). Cannot commit." % (sLocalVersion, sRemoteVersion)
                )
            else:
                gLogger.info("AutoMerging new data!")
                if updateVersionOption:
                    return S_ERROR("Cannot AutoMerge! version was overwritten")
                result = self.__mergeIndependentUpdates(oRemoteConfData)
                if not result["OK"]:
                    gLogger.warn("Could not AutoMerge!", result["Message"])
                    return S_ERROR("AutoMerge failed: %s" % result["Message"])
                requestedRemoteCFG = result["Value"]
                gLogger.info("AutoMerge successful!")
                oRemoteConfData.setRemoteCFG(requestedRemoteCFG)
        # Test that configuration names are the same
        sRemoteName = oRemoteConfData.getName()
        sLocalName = gConfigurationData.getName()
        if sRemoteName != sLocalName:
            return S_ERROR("Names differ: Server is %s and remote is %s" % (sLocalName, sRemoteName))
        # Update and generate a new version
        gLogger.info("Committing new data...")
        gConfigurationData.lock()
        gLogger.info("Setting the new CFG")
        gConfigurationData.setRemoteCFG(oRemoteConfData.getRemoteCFG())
        gConfigurationData.unlock()
        gLogger.info("Generating new version")
        gConfigurationData.generateNewVersion()
        # self.__checkSlavesStatus( forceWriteConfiguration = True )
        gLogger.info("Writing new version to disk")
        retVal = gConfigurationData.writeRemoteConfigurationToDisk(
            "%s@%s" % (committer, gConfigurationData.getVersion())
        )
        gLogger.info("New version", gConfigurationData.getVersion())

        # Attempt to update the configuration on currently registered slave services
        if gConfigurationData.getAutoSlaveSync():
            self.forceSlavesUpdate()

        return retVal

    def getCompressedConfigurationData(self):
        return gConfigurationData.getCompressedData()

    def getVersion(self):
        return gConfigurationData.getVersion()

    def getCommitHistory(self):
        files = self.__getCfgBackups(gConfigurationData.getBackupDir())
        backups = [".".join(fileName.split(".")[1:-1]).split("@") for fileName in files]
        return backups

    def getVersionContents(self, date):
        backupDir = gConfigurationData.getBackupDir()
        files = self.__getCfgBackups(backupDir, date)
        for fileName in files:
            with zipfile.ZipFile("%s/%s" % (backupDir, fileName), "rb") as zFile:
                cfgName = zFile.namelist()[0]
                retVal = S_OK(zlib.compress(zFile.read(cfgName), 9))
            return retVal
        return S_ERROR("Version %s does not exist" % date)

    def __getCfgBackups(self, basePath, date="", subPath=""):
        rs = re.compile(r"^%s\..*%s.*\.zip$" % (gConfigurationData.getName(), date))
        fsEntries = os.listdir("%s/%s" % (basePath, subPath))
        fsEntries.sort(reverse=True)
        backupsList = []
        for entry in fsEntries:
            entryPath = "%s/%s/%s" % (basePath, subPath, entry)
            if os.path.isdir(entryPath):
                backupsList.extend(self.__getCfgBackups(basePath, date, "%s/%s" % (subPath, entry)))
            elif os.path.isfile(entryPath):
                if rs.search(entry):
                    backupsList.append("%s/%s" % (subPath, entry))
        return backupsList

    def __getPreviousCFG(self, oRemoteConfData):
        backupsList = self.__getCfgBackups(gConfigurationData.getBackupDir(), date=oRemoteConfData.getVersion())
        if not backupsList:
            return S_ERROR("Could not AutoMerge. Could not retrieve original committer's version")
        prevRemoteConfData = ConfigurationData()
        backFile = backupsList[0]
        if backFile[0] == "/":
            backFile = os.path.join(gConfigurationData.getBackupDir(), backFile[1:])
        try:
            prevRemoteConfData.loadConfigurationData(backFile)
        except Exception as e:
            return S_ERROR("Could not load original committer's version: %s" % str(e))
        gLogger.info("Loaded client original version %s" % prevRemoteConfData.getVersion())
        return S_OK(prevRemoteConfData.getRemoteCFG())

    def _checkConflictsInModifications(self, realModList, reqModList, parentSection=""):
        realModifiedSections = dict(
            [(modAc[1], modAc[3]) for modAc in realModList if modAc[0].find("Sec") == len(modAc[0]) - 3]
        )
        reqOptionsModificationList = dict(
            [(modAc[1], modAc[3]) for modAc in reqModList if modAc[0].find("Opt") == len(modAc[0]) - 3]
        )
        for modAc in reqModList:
            action = modAc[0]
            objectName = modAc[1]
            if action == "addSec":
                if objectName in realModifiedSections:
                    return S_ERROR("Section %s/%s already exists" % (parentSection, objectName))
            elif action == "delSec":
                if objectName in realModifiedSections:
                    return S_ERROR(
                        "Section %s/%s cannot be deleted. It has been modified." % (parentSection, objectName)
                    )
            elif action == "modSec":
                if objectName in realModifiedSections:
                    result = self._checkConflictsInModifications(
                        realModifiedSections[objectName], modAc[3], "%s/%s" % (parentSection, objectName)
                    )
                    if not result["OK"]:
                        return result
        for modAc in realModList:
            action = modAc[0]
            objectName = modAc[1]
            if action.find("Opt") == len(action) - 3:
                return S_ERROR(
                    "Section %s cannot be merged. Option %s/%s has been modified"
                    % (parentSection, parentSection, objectName)
                )
        return S_OK()

    def __mergeIndependentUpdates(self, oRemoteConfData):
        # return S_ERROR( "AutoMerge is still not finished.
        # Meanwhile... why don't you get the newest conf and update from there?" )
        # Get all the CFGs
        curSrvCFG = gConfigurationData.getRemoteCFG().clone()
        curCliCFG = oRemoteConfData.getRemoteCFG().clone()
        result = self.__getPreviousCFG(oRemoteConfData)
        if not result["OK"]:
            return result
        prevCliCFG = result["Value"]
        # Try to merge curCli with curSrv. To do so we check the updates from
        # prevCli -> curSrv VS prevCli -> curCli
        prevCliToCurCliModList = prevCliCFG.getModifications(curCliCFG)
        prevCliToCurSrvModList = prevCliCFG.getModifications(curSrvCFG)
        result = self._checkConflictsInModifications(prevCliToCurSrvModList, prevCliToCurCliModList)
        if not result["OK"]:
            return S_ERROR("Cannot AutoMerge: %s" % result["Message"])
        # Merge!
        result = curSrvCFG.applyModifications(prevCliToCurCliModList)
        if not result["OK"]:
            return result
        return S_OK(curSrvCFG)



""" Module that contains simple client access to Matcher service
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.Core.Utilities.JEncode import strToIntDict


@createClient("WorkloadManagement/Matcher")
class MatcherClient(Client):

    """Exposes the functionality available in the WorkloadManagement/MatcherHandler

    This inherits the DIRAC base Client for direct execution of server functionality.
    The following methods are available (although not visible here).

    """

    def __init__(self, **kwargs):
        """Simple constructor"""

        super(MatcherClient, self).__init__(**kwargs)
        self.setServer("WorkloadManagement/Matcher")

    @ignoreEncodeWarning
    def getMatchingTaskQueues(self, resourceDict):
        """Return all task queues that match the resourceDict"""
        res = self._getRPC().getMatchingTaskQueues(resourceDict)

        if res["OK"]:
            # Cast the string back to int
            res["Value"] = strToIntDict(res["Value"])
        return res

    @ignoreEncodeWarning
    def getActiveTaskQueues(self):
        """Return all active task queues"""
        res = self._getRPC().getActiveTaskQueues()

        if res["OK"]:
            # Cast the string back to int
            res["Value"] = strToIntDict(res["Value"])
        return res



#########################################################################################
# Torque.py
# 10.11.2014
# Author: A.T.
#########################################################################################

""" Torque.py is a DIRAC independent class representing Torque batch system.
    Torque objects are used as backend batch system representation for
    LocalComputingElement and SSHComputingElement classes
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import subprocess
import shlex
import os

__RCSID__ = "$Id$"


class Torque(object):
    def submitJob(self, **kwargs):
        """Submit nJobs to the Torque batch system"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["Executable", "OutputDir", "ErrorDir", "Queue", "SubmitOptions"]

        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        nJobs = kwargs.get("NJobs")
        if not nJobs:
            nJobs = 1

        jobIDs = []
        status = -1

        preamble = kwargs.get("Preamble")
        for _i in range(int(nJobs)):
            cmd = "%s; " % preamble if preamble else ""
            cmd += (
                "qsub -o %(OutputDir)s "
                "-e %(ErrorDir)s "
                "-q %(Queue)s "
                "-N DIRACPilot "
                "%(SubmitOptions)s %(Executable)s 2>/dev/null" % kwargs
            )
            sp = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status == 0:
                jobIDs.append(output.split(".")[0])
            else:
                break

        if jobIDs:
            resultDict["Status"] = 0
            resultDict["Jobs"] = jobIDs
        else:
            resultDict["Status"] = status
            resultDict["Message"] = error
        return resultDict

    def getJobStatus(self, **kwargs):
        """Get the status information for the given list of jobs"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs["JobIDList"]
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        jobDict = {}
        for job in jobIDList:
            if not job:
                continue
            jobNumber = job
            jobDict[jobNumber] = job

        cmd = "qstat " + " ".join(jobIDList)
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        statusDict = {}
        output = output.replace("\r", "")
        lines = output.split("\n")
        for job in jobDict:
            statusDict[jobDict[job]] = "Unknown"
            for line in lines:
                if line.find(job) != -1:
                    if line.find("Unknown") != -1:
                        statusDict[jobDict[job]] = "Unknown"
                    else:
                        torqueStatus = line.split()[4]
                        if torqueStatus in ["E"]:
                            statusDict[jobDict[job]] = "Done"
                        elif torqueStatus in ["R", "C"]:
                            statusDict[jobDict[job]] = "Running"
                        elif torqueStatus in ["S", "W", "Q", "H", "T"]:
                            statusDict[jobDict[job]] = "Waiting"

        # Final output
        status = 0
        resultDict["Status"] = 0
        resultDict["Jobs"] = statusDict
        return resultDict

    def getCEStatus(self, **kwargs):
        """Get the overall CE status"""

        resultDict = {}

        user = kwargs.get("User")
        if not user:
            user = os.environ.get("USER")
        if not user:
            resultDict["Status"] = -1
            resultDict["Message"] = "No user name"
            return resultDict

        cmd = "qselect -u %s -s WQ | wc -l; qselect -u %s -s R | wc -l" % (user, user)
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        waitingJobs, runningJobs = output.split()[:2]

        # Final output
        try:
            resultDict["Status"] = 0
            resultDict["Waiting"] = int(waitingJobs)
            resultDict["Running"] = int(runningJobs)
        except Exception as e:
            resultDict["Status"] = -1
            resultDict["Output"] = output
            resultDict["Message"] = "Exception: %s" % str(e)

        return resultDict

    def killJob(self, **kwargs):
        """Kill all jobs in the given list"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs["JobIDList"]
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        successful = []
        failed = []
        errors = ""
        for job in jobIDList:
            sp = subprocess.Popen(
                shlex.split("qdel %s" % job),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status != 0:
                failed.append(job)
                errors += error
            else:
                successful.append(job)

        resultDict["Status"] = 0
        if failed:
            resultDict["Status"] = 1
            resultDict["Message"] = errors
        resultDict["Successful"] = successful
        resultDict["Failed"] = failed
        return resultDict

    def getJobOutputFiles(self, **kwargs):
        """Get output file names and templates for the specific CE"""
        resultDict = {}
        MANDATORY_PARAMETERS = ["JobIDList", "OutputDir", "ErrorDir"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        outputDir = kwargs["OutputDir"]
        errorDir = kwargs["ErrorDir"]

        outputTemplate = "%s/DIRACPilot.o%%s" % outputDir
        errorTemplate = "%s/DIRACPilot.e%%s" % errorDir
        outputTemplate = os.path.expandvars(outputTemplate)
        errorTemplate = os.path.expandvars(errorTemplate)

        jobIDList = kwargs["JobIDList"]

        jobDict = {}
        for job in jobIDList:
            jobDict[job] = {}
            jobDict[job]["Output"] = outputTemplate % job
            jobDict[job]["Error"] = errorTemplate % job

        resultDict["Status"] = 0
        resultDict["Jobs"] = jobDict
        resultDict["OutputTemplate"] = outputTemplate
        resultDict["ErrorTemplate"] = errorTemplate

        return resultDict



""" :mod: DMSRequestOperationsBase

    ====================

    Just a collector of common functions
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

from DIRAC import S_OK, S_ERROR

from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers


class DMSRequestOperationsBase(OperationHandlerBase):
    def __init__(self, operation=None, csPath=None):
        OperationHandlerBase.__init__(self, operation, csPath)
        self.registrationProtocols = DMSHelpers().getRegistrationProtocols()

    def checkSEsRSS(self, checkSEs=None, access="WriteAccess", failIfBanned=True):
        """check SEs.
        By default, we check the SEs for WriteAccess, but it is configurable
        """
        if not checkSEs:
            checkSEs = self.operation.targetSEList
        elif isinstance(checkSEs, str):
            checkSEs = [checkSEs]

        if access == "ReadAccess":
            seType = "sourceSE"
        else:
            seType = "targetSE"
        bannedSEs = []

        for checkSE in checkSEs:
            seStatus = self.rssSEStatus(checkSE, access, retries=5)
            if not seStatus["OK"]:
                self.log.error("Failed to get SE status", seStatus["Message"])
                error = "unknown %s: %s" % (seType, checkSE)
                for opFile in self.operation:
                    opFile.Error = error
                self.operation.Error = error
                return S_ERROR(error)

            if not seStatus["Value"]:
                self.log.info("%s %s is banned for %s right now" % (seType.capitalize(), checkSE, access))
                bannedSEs.append(checkSE)
                self.operation.Error = "banned %s: %s;" % (seType, checkSE)

        if bannedSEs:
            alwaysBannedSEs = []
            for seName in bannedSEs:
                res = self.rssClient().isStorageElementAlwaysBanned(seName, access)
                if not res["OK"]:
                    continue

                # The SE will always be banned
                if res["Value"]:
                    alwaysBannedSEs.append(seName)

            # If Some SE are always banned, we fail the request
            if alwaysBannedSEs:
                self.operation.Error = "%s always banned" % alwaysBannedSEs
                if failIfBanned:
                    self.log.info("Some storages are always banned, failing the request", alwaysBannedSEs)
                    for opFile in self.operation:
                        opFile.Error = "%s always banned" % alwaysBannedSEs
                        opFile.Status = "Failed"

            # If it is temporary, we wait an hour
            else:
                self.log.info("Banning is temporary, next attempt in an hour")
                self.operation.Error = "%s currently banned" % bannedSEs
                self.request.delayNextExecution(60)

        return S_OK(bannedSEs)

    def getRegisterOperation(self, opFile, targetSE, type="RegisterFile", catalog=None):
        """add RegisterReplica operation for file

        :param ~DIRAC.RequestManagementSystem.Client.File.File opFile: operation file
        :param str targetSE: target SE
        """
        # # add RegisterReplica operation
        registerOperation = Operation()
        registerOperation.Type = type
        registerOperation.TargetSE = targetSE
        if catalog:
            registerOperation.Catalog = catalog

        registerFile = File()
        registerFile.LFN = opFile.LFN
        registerFile.PFN = (
            StorageElement(targetSE)
            .getURL(opFile.LFN, protocol=self.registrationProtocols)
            .get("Value", {})
            .get("Successful", {})
            .get(opFile.LFN)
        )
        registerFile.GUID = opFile.GUID
        registerFile.Checksum = opFile.Checksum
        registerFile.ChecksumType = opFile.ChecksumType
        registerFile.Size = opFile.Size

        registerOperation.addFile(registerFile)
        return registerOperation



"""
  TornadoClient is equivalent of the RPCClient but in HTTPS.
  Usage of TornadoClient is the same as RPCClient, you can instantiate TornadoClient with
  complete url (https://host:port/System/Component) or just "System/Component". Like RPCClient
  you can use all method defined in your service, your call will be automatically transformed
  in RPC.

  It also exposes the same interface for receiving file as the TransferClient.

  Main changes:

  - KeepAliveLapse is removed, requests library manages it itself.
  - nbOfRetry (defined as private attribute) is removed, requests library manage it itself.
  - Underneath it uses HTTP POST protocol and JSON. See  :ref:`httpsTornado` for details

  Example::

    from DIRAC.Core.Tornado.Client.TornadoClient import TornadoClient
    myService = TornadoClient("Framework/MyService")
    myService.doSomething() #Returns S_OK/S_ERROR

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# pylint: disable=broad-except

from DIRAC.Core.Tornado.Client.private.TornadoBaseClient import TornadoBaseClient
from DIRAC.Core.Utilities.JEncode import encode


class TornadoClient(TornadoBaseClient):
    """
    Client for calling tornado services
    Interface is based on RPCClient interface
    """

    def __getattr__(self, attrname):
        """
        Return the RPC call procedure

        :param str attrname: Name of the procedure we are trying to call
        :return: RPC procedure as function
        """

        def call(*args):
            """
            Just returns the right function for RPC Call
            """
            return self.executeRPC(attrname, *args)

        return call

    # Name from RPCClient Interface
    def executeRPC(self, method, *args):
        """
        Calls a remote service

        :param str method: remote procedure name
        :param args: list of arguments
        :returns: decoded response from server, server may return S_OK or S_ERROR
        """
        rpcCall = {"method": method, "args": encode(args)}
        # Start request
        retVal = self._request(**rpcCall)
        retVal["rpcStub"] = (self._getBaseStub(), method, list(args))
        return retVal

    def receiveFile(self, destFile, *args):
        """
        Equivalent of :py:meth:`~DIRAC.Core.DISET.TransferClient.TransferClient.receiveFile`

        In practice, it calls the remote method `streamToClient` and stores the raw result in a file

        :param str destFile: path where to store the result
        :param args: list of arguments
        :returns: S_OK/S_ERROR
        """
        rpcCall = {"method": "streamToClient", "args": encode(args)}
        # Start request
        retVal = self._request(outputFile=destFile, **rpcCall)
        return retVal


def executeRPCStub(rpcStub):
    """
    Playback a stub
    # Copy-paste from DIRAC.Core.DISET.RPCClient with RPCClient changed into TornadoClient
    """
    # Generate a client with the same parameters
    client = TornadoClient(rpcStub[0][0], **rpcStub[0][1])
    # Get a functor to execute the RPC call
    rpcFunc = getattr(client, rpcStub[1])
    # Reproduce the call
    return rpcFunc(*rpcStub[2])



#! /usr/bin/env python

# FIXME: if it requires a dirac.cfg it is not a unit test and should be moved to tests directory


from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import unittest
import time
import os
import shutil
import sys
import six

from DIRAC.Core.Base.Script import parseCommandLine, getPositionalArgs

parseCommandLine()

from DIRAC.Resources.Storage.StorageFactory import StorageFactory
from DIRAC.Core.Utilities.File import getSize

positionalArgs = getPositionalArgs()

__RCSID__ = "$Id$"

if len(positionalArgs) < 2:
    print("Usage: TestStoragePlugIn.py StorageElement plugin")
    sys.exit()
else:
    storageElementToTest = positionalArgs[0]
    plugin = positionalArgs[1]


class StoragePlugInTestCase(unittest.TestCase):
    """Base class for the StoragePlugin test cases"""

    def setUp(self):

        factory = StorageFactory("lhcb")
        res = factory.getStorages(storageElementToTest, [plugin])
        self.assertTrue(res["OK"])
        storageDetails = res["Value"]
        self.storage = storageDetails["StorageObjects"][0]
        self.storage.changeDirectory("lhcb/test/unit-test/TestStoragePlugIn")
        destDir = self.storage.getCurrentURL("")["Value"]
        res = self.storage.createDirectory(destDir)
        self.assertTrue(res["OK"])
        self.assertTrue(destDir in res["Value"]["Successful"])
        self.assertTrue(res["Value"]["Successful"][destDir])
        self.numberOfFiles = 1

    def tearDown(self):
        remoteDir = self.storage.getCurrentURL("")["Value"]
        _ = self.storage.removeDirectory(remoteDir, True)


class DirectoryTestCase(StoragePlugInTestCase):
    def test_putRemoveDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tPut Directory test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("")["Value"]
        _ = self.storage.removeDirectory(remoteDir, True)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)

        # Check that we can successfully upload the directory to the storage element
        dirDict = {remoteDir: localDir}
        putDirRes = self.storage.putDirectory(dirDict)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir, True)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        if putDirRes["Value"]["Successful"][remoteDir]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Files"], self.numberOfFiles)
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Size"], self.numberOfFiles * sizeOfLocalFile)
        self.assertTrue(isinstance(putDirRes["Value"]["Successful"][remoteDir]["Files"], int))
        self.assertTrue(type(putDirRes["Value"]["Successful"][remoteDir]["Size"]) in six.integer_types)
        # Perform the checks for the remove dir operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        if removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
        self.assertTrue(isinstance(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], int))
        self.assertTrue(type(removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"]) in six.integer_types)

    def test_isDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tIs Directory test\n"
        )
        # Test that we can determine what is a directory
        destDir = self.storage.getCurrentURL("")["Value"]
        isDirRes = self.storage.isDirectory(destDir)
        # Test that we can determine that a directory is not a directory
        dummyDir = self.storage.getCurrentURL("NonExistantFile")["Value"]
        nonExistantDirRes = self.storage.isDirectory(dummyDir)

        # Check the is directory operation
        self.assertTrue(isDirRes["OK"])
        self.assertTrue(destDir in isDirRes["Value"]["Successful"])
        self.assertTrue(isDirRes["Value"]["Successful"][destDir])
        # Check the non existant directory operation
        self.assertTrue(nonExistantDirRes["OK"])
        self.assertTrue(dummyDir in nonExistantDirRes["Value"]["Failed"])
        expectedError = "Path does not exist"
        self.assertTrue(expectedError in nonExistantDirRes["Value"]["Failed"][dummyDir])

    def test_putGetDirectoryMetadata(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet Directory Metadata test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("")["Value"]
        _ = self.storage.removeDirectory(remoteDir, True)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)

        # Check that we can successfully upload the directory to the storage element
        dirDict = {remoteDir: localDir}
        putDirRes = self.storage.putDirectory(dirDict)

        # Get the directory metadata
        getMetadataRes = self.storage.getDirectoryMetadata(remoteDir)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir, True)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        if putDirRes["Value"]["Successful"][remoteDir]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Files"], self.numberOfFiles)
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Size"], self.numberOfFiles * sizeOfLocalFile)
        self.assertTrue(isinstance(putDirRes["Value"]["Successful"][remoteDir]["Files"], int))
        self.assertTrue(type(putDirRes["Value"]["Successful"][remoteDir]["Size"]) in six.integer_types)
        # Perform the checks for the get metadata operation
        self.assertTrue(getMetadataRes["OK"])
        self.assertTrue(remoteDir in getMetadataRes["Value"]["Successful"])
        resDict = getMetadataRes["Value"]["Successful"][remoteDir]
        self.assertTrue("Mode" in resDict)
        self.assertTrue(isinstance(resDict["Mode"], int))
        # Perform the checks for the remove directory operation
        if removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
        self.assertTrue(isinstance(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], int))
        self.assertTrue(type(removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"]) in six.integer_types)

    def test_putGetDirectorySize(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet Directory Size test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("")["Value"]
        _ = self.storage.removeDirectory(remoteDir, True)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirDict = {remoteDir: localDir}
        putDirRes = self.storage.putDirectory(dirDict)
        # Now get the directory size
        getDirSizeRes = self.storage.getDirectorySize(remoteDir)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir, True)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        if putDirRes["Value"]["Successful"][remoteDir]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Files"], self.numberOfFiles)
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Size"], self.numberOfFiles * sizeOfLocalFile)
        self.assertTrue(isinstance(putDirRes["Value"]["Successful"][remoteDir]["Files"], int))
        self.assertTrue(type(putDirRes["Value"]["Successful"][remoteDir]["Size"]) in six.integer_types)
        # Now perform the checks for the get directory size operation
        self.assertTrue(getDirSizeRes["OK"])
        self.assertTrue(remoteDir in getDirSizeRes["Value"]["Successful"])
        resDict = getDirSizeRes["Value"]["Successful"][remoteDir]
        self.assertTrue(type(resDict["Size"]) in six.integer_types)
        self.assertTrue(isinstance(resDict["Files"], int))
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        if removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
        self.assertTrue(isinstance(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], int))
        self.assertTrue(type(removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"]) in six.integer_types)

    def test_putListDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tList Directory test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("")["Value"]
        _ = self.storage.removeDirectory(remoteDir, True)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirDict = {remoteDir: localDir}
        putDirRes = self.storage.putDirectory(dirDict)
        # List the remote directory
        listDirRes = self.storage.listDirectory(remoteDir)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir, True)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        if putDirRes["Value"]["Successful"][remoteDir]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Files"], self.numberOfFiles)
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Size"], self.numberOfFiles * sizeOfLocalFile)
        self.assertTrue(isinstance(putDirRes["Value"]["Successful"][remoteDir]["Files"], int))
        self.assertTrue(type(putDirRes["Value"]["Successful"][remoteDir]["Size"]) in six.integer_types)
        # Perform the checks for the list dir operation
        self.assertTrue(listDirRes["OK"])
        self.assertTrue(remoteDir in listDirRes["Value"]["Successful"])
        resDict = listDirRes["Value"]["Successful"][remoteDir]
        self.assertTrue("SubDirs" in resDict)
        self.assertTrue("Files" in resDict)
        self.assertEqual(len(resDict["Files"]), self.numberOfFiles)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        if removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
        self.assertTrue(isinstance(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], int))
        self.assertTrue(type(removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"]) in six.integer_types)

    def test_putGetDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet Directory test\n"
        )
        # First clean the remote directory incase something was left there
        remoteDir = self.storage.getCurrentURL("")["Value"]
        _ = self.storage.removeDirectory(remoteDir, True)

        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirDict = {remoteDir: localDir}
        putDirRes = self.storage.putDirectory(dirDict)
        # Clean up the locally created directory
        shutil.rmtree(localDir)
        # Check that we can get directories from the storage element
        getDirRes = self.storage.getDirectory(remoteDir, localPath=localDir)
        # Now remove the remove directory
        removeDirRes = self.storage.removeDirectory(remoteDir, True)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(remoteDir in putDirRes["Value"]["Successful"])
        if putDirRes["Value"]["Successful"][remoteDir]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Files"], self.numberOfFiles)
            self.assertEqual(putDirRes["Value"]["Successful"][remoteDir]["Size"], self.numberOfFiles * sizeOfLocalFile)
        self.assertTrue(isinstance(putDirRes["Value"]["Successful"][remoteDir]["Files"], int))
        self.assertTrue(type(putDirRes["Value"]["Successful"][remoteDir]["Size"]) in six.integer_types)
        # Perform the checks for the get dir operation
        self.assertTrue(getDirRes["OK"])
        self.assertTrue(remoteDir in getDirRes["Value"]["Successful"])
        resDict = getDirRes["Value"]["Successful"][remoteDir]
        if resDict["Files"]:
            self.assertEqual(resDict["Files"], self.numberOfFiles)
            self.assertEqual(resDict["Size"], self.numberOfFiles * sizeOfLocalFile)
        self.assertTrue(isinstance(resDict["Files"], int))
        self.assertTrue(type(resDict["Size"]) in six.integer_types)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(remoteDir in removeDirRes["Value"]["Successful"])
        if removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
        self.assertTrue(isinstance(removeDirRes["Value"]["Successful"][remoteDir]["FilesRemoved"], int))
        self.assertTrue(type(removeDirRes["Value"]["Successful"][remoteDir]["SizeRemoved"]) in six.integer_types)


class FileTestCase(StoragePlugInTestCase):
    def test_putRemoveFile(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tPut and Remove test\n"
        )

        # Make sure that we can actually upload a file properly
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        destFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {destFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Make sure we are able to remove the file
        removeFileRes = self.storage.removeFile(destFile)

        # Check the successful put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(destFile in putFileRes["Value"]["Successful"])
        self.assertEqual(putFileRes["Value"]["Successful"][destFile], srcFileSize)
        # Check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(destFile in removeFileRes["Value"]["Successful"])
        self.assertTrue(removeFileRes["Value"]["Successful"][destFile])

    def test_putGetFile(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tPut and Get test\n"
        )

        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)

        # Then make sure we can get a local copy of the file
        getFileRes = self.storage.getFile(remoteFile)
        # Cleanup the remote mess
        removeFileRes = self.storage.removeFile(remoteFile)
        # Cleanup the mess locally
        os.remove(testFileName)

        # Check the put operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check the get operation
        self.assertTrue(getFileRes["OK"])
        self.assertTrue(remoteFile in getFileRes["Value"]["Successful"])
        self.assertEqual(getFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check the remove operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])
        self.assertTrue(removeFileRes["Value"]["Successful"][remoteFile])

    def test_putExistsFile(self):
        print("\n\n#########################################################" "################\n\n\t\t\tExists test\n")
        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Then get the file's existance
        existsFileRes = self.storage.exists(remoteFile)
        # Now remove the file
        removeFileRes = self.storage.removeFile(remoteFile)
        # Check  again that the file exists
        failedExistRes = self.storage.exists(remoteFile)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check the exists operation
        self.assertTrue(existsFileRes["OK"])
        self.assertTrue(remoteFile in existsFileRes["Value"]["Successful"])
        self.assertTrue(existsFileRes["Value"]["Successful"][remoteFile])
        # Check the removal operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])
        # Check the failed exists operation
        self.assertTrue(failedExistRes["OK"])
        self.assertTrue(remoteFile in failedExistRes["Value"]["Successful"])
        self.assertFalse(failedExistRes["Value"]["Successful"][remoteFile])

    def test_putIsFile(self):
        print(
            "\n\n#########################################################" "################\n\n\t\t\tIs file test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Check we are able to determine that it is a file
        isFileRes = self.storage.isFile(remoteFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(remoteFile)
        # Check that everything isn't a file
        remoteDir = os.path.dirname(remoteFile)
        failedIsFileRes = self.storage.isFile(remoteDir)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertTrue(putFileRes["Value"]["Successful"][remoteFile])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check the is file operation
        self.assertTrue(isFileRes["OK"])
        self.assertTrue(remoteFile in isFileRes["Value"]["Successful"])
        self.assertTrue(isFileRes["Value"]["Successful"][remoteFile])
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])
        # Check that the directory is not a file
        self.assertTrue(failedIsFileRes["OK"])
        self.assertTrue(remoteDir in failedIsFileRes["Value"]["Successful"])
        self.assertFalse(failedIsFileRes["Value"]["Successful"][remoteDir])

    def test_putGetFileMetaData(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet file metadata test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Check that we can get the file metadata
        getMetadataRes = self.storage.getFileMetadata(remoteFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(remoteFile)
        # See what happens with non existant files
        failedMetadataRes = self.storage.getFileMetadata(remoteFile)
        # Check directories are handled properly
        remoteDir = os.path.dirname(remoteFile)
        directoryMetadataRes = self.storage.getFileMetadata(remoteDir)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertTrue(putFileRes["Value"]["Successful"][remoteFile])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check the get metadata operation
        self.assertTrue(getMetadataRes["OK"])
        self.assertTrue(remoteFile in getMetadataRes["Value"]["Successful"])
        fileMetaData = getMetadataRes["Value"]["Successful"][remoteFile]
        self.assertTrue(fileMetaData["Cached"])
        self.assertFalse(fileMetaData["Migrated"])
        self.assertEqual(fileMetaData["Size"], srcFileSize)
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])
        # Check the get metadata for non existant file
        self.assertTrue(failedMetadataRes["OK"])
        self.assertTrue(remoteFile in failedMetadataRes["Value"]["Failed"])
        expectedError = "File does not exist"
        self.assertTrue(expectedError in failedMetadataRes["Value"]["Failed"][remoteFile])
        # Check that metadata operation with a directory
        self.assertTrue(directoryMetadataRes["OK"])
        self.assertTrue(remoteDir in directoryMetadataRes["Value"]["Failed"])
        expectedError = "Supplied path is not a file"
        self.assertTrue(expectedError in directoryMetadataRes["Value"]["Failed"][remoteDir])

    def test_putGetFileSize(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet file size test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Check that we can get the file size
        getSizeRes = self.storage.getFileSize(remoteFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(remoteFile)
        # Check non existant files
        failedSizeRes = self.storage.getFileMetadata(remoteFile)
        # Check directories are handled properly
        remoteDir = os.path.dirname(remoteFile)
        directorySizeRes = self.storage.getFileSize(remoteDir)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertTrue(putFileRes["Value"]["Successful"][remoteFile])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check that we got the file size correctly
        self.assertTrue(getSizeRes["OK"])
        self.assertTrue(remoteFile in getSizeRes["Value"]["Successful"])
        self.assertEqual(getSizeRes["Value"]["Successful"][remoteFile], srcFileSize)
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])
        # Check the get size with non existant file works properly
        self.assertTrue(failedSizeRes["OK"])
        self.assertTrue(remoteFile in failedSizeRes["Value"]["Failed"])
        expectedError = "File does not exist"
        self.assertTrue(expectedError in failedSizeRes["Value"]["Failed"][remoteFile])
        # Check that the passing a directory is handled correctly
        self.assertTrue(directorySizeRes["OK"])
        self.assertTrue(remoteDir in directorySizeRes["Value"]["Failed"])
        expectedError = "Supplied path is not a file"
        self.assertTrue(expectedError in directorySizeRes["Value"]["Failed"][remoteDir])

    def test_putPrestageFile(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tFile prestage test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Check that we can issue a stage request
        prestageRes = self.storage.prestageFile(remoteFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(remoteFile)
        # Check what happens with deleted files
        deletedPrestageRes = self.storage.prestageFile(remoteFile)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertTrue(putFileRes["Value"]["Successful"][remoteFile])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check the prestage file operation
        self.assertTrue(prestageRes["OK"])
        self.assertTrue(remoteFile in prestageRes["Value"]["Successful"])
        self.assertTrue(prestageRes["Value"]["Successful"][remoteFile])
        # Check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])
        # Check that pre-staging non-existant file fails
        self.assertTrue(deletedPrestageRes["OK"])
        self.assertTrue(remoteFile in deletedPrestageRes["Value"]["Failed"])
        expectedError = "No such file or directory"
        self.assertTrue(expectedError in deletedPrestageRes["Value"]["Failed"][remoteFile])

    def test_putFilegetTransportURL(self):
        print(
            "\n\n#########################################################" "################\n\n\t\t\tGet tURL test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Check that we can get a turl
        getTurlRes = self.storage.getTransportURL(remoteFile)
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(remoteFile)
        # Try and get a turl for a non existant file
        failedGetTurlRes = self.storage.getTransportURL(remoteFile)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertTrue(putFileRes["Value"]["Successful"][remoteFile])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # check the get turl operation
        self.assertTrue(getTurlRes["OK"])
        self.assertTrue(remoteFile in getTurlRes["Value"]["Successful"])
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])
        # Check the get turl with non existant file operation
        self.assertTrue(failedGetTurlRes["OK"])
        self.assertTrue(remoteFile in failedGetTurlRes["Value"]["Failed"])
        expectedError = "File does not exist"
        self.assertTrue(expectedError in failedGetTurlRes["Value"]["Failed"][remoteFile])

    def test_putPinRelease(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tPin and Release test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Check that we can pin the file
        pinFileRes = self.storage.pinFile(remoteFile)
        srmID = ""
        if pinFileRes["OK"]:
            if remoteFile in pinFileRes["Value"]["Successful"]:
                srmID = pinFileRes["Value"]["Successful"][remoteFile]
        # Check that we can release the file
        releaseFileRes = self.storage.releaseFile({remoteFile: srmID})
        # Clean up the mess
        removeFileRes = self.storage.removeFile(remoteFile)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertTrue(putFileRes["Value"]["Successful"][remoteFile])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check the pin file operation
        self.assertTrue(pinFileRes["OK"])
        self.assertTrue(remoteFile in pinFileRes["Value"]["Successful"])
        self.assertTrue(type(pinFileRes["Value"]["Successful"][remoteFile]) in six.string_types)
        # Check the release file operation
        self.assertTrue(releaseFileRes["OK"])
        self.assertTrue(remoteFile in releaseFileRes["Value"]["Successful"])
        # check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])

    def test_putPrestageStatus(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tPrestage status test\n"
        )
        # First upload a file to the storage
        srcFile = "/etc/group"
        srcFileSize = getSize(srcFile)
        testFileName = "testFile.%s" % time.time()
        remoteFile = self.storage.getCurrentURL(testFileName)["Value"]
        fileDict = {remoteFile: srcFile}
        putFileRes = self.storage.putFile(fileDict)
        # Check that we can issue a stage request
        prestageRes = self.storage.prestageFile(remoteFile)
        srmID = ""
        if prestageRes["OK"]:
            if remoteFile in prestageRes["Value"]["Successful"]:
                srmID = prestageRes["Value"]["Successful"][remoteFile]
        # Take a quick break to allow the SRM to realise the file is available
        sleepTime = 10
        print("Sleeping for %s seconds" % sleepTime)
        time.sleep(sleepTime)
        # Check that we can monitor the stage request
        prestageStatusRes = self.storage.prestageFileStatus({remoteFile: srmID})
        # Clean up the remote mess
        removeFileRes = self.storage.removeFile(remoteFile)

        # Check the put file operation
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(remoteFile in putFileRes["Value"]["Successful"])
        self.assertTrue(putFileRes["Value"]["Successful"][remoteFile])
        self.assertEqual(putFileRes["Value"]["Successful"][remoteFile], srcFileSize)
        # Check the prestage file operation
        self.assertTrue(prestageRes["OK"])
        self.assertTrue(remoteFile in prestageRes["Value"]["Successful"])
        self.assertTrue(prestageRes["Value"]["Successful"][remoteFile])
        self.assertTrue(type(prestageRes["Value"]["Successful"][remoteFile]) in six.string_types)
        # Check the prestage status operation
        self.assertTrue(prestageStatusRes["OK"])
        self.assertTrue(remoteFile in prestageStatusRes["Value"]["Successful"])
        self.assertTrue(prestageStatusRes["Value"]["Successful"][remoteFile])
        # Check the remove file operation
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(remoteFile in removeFileRes["Value"]["Successful"])


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(FileTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DirectoryTestCase))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



""" SiteStatus helper

Module that acts as a helper for knowing the status of a site.
It takes care of switching between the CS and the RSS.
The status is kept in the RSSCache object, which is a small wrapper on top of DictCache

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import errno
import math
from time import sleep
from datetime import datetime, timedelta

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.Client.WMSAdministratorClient import WMSAdministratorClient
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
from DIRAC.ResourceStatusSystem.Utilities.RSSCacheNoThread import RSSCache
from DIRAC.ResourceStatusSystem.Utilities.RssConfiguration import RssConfiguration


@six.add_metaclass(DIRACSingleton)
class SiteStatus(object):
    """
    RSS helper to interact with the 'Site' family on the DB. It provides the most
    demanded functions and a cache to avoid hitting the server too often.

    It provides four methods to interact with the site statuses:
    * getSiteStatuses
    * isUsableSite
    * getUsableSites
    * getSites
    """

    def __init__(self):
        """
        Constructor, initializes the rssClient.
        """

        self.log = gLogger.getSubLogger(self.__class__.__name__)
        self.rssConfig = RssConfiguration()
        self.__opHelper = Operations()
        self.rssFlag = ResourceStatus().rssFlag
        self.rsClient = ResourceStatusClient()

        cacheLifeTime = int(self.rssConfig.getConfigCache())

        # RSSCache only affects the calls directed to RSS, if using the CS it is not used.
        self.rssCache = RSSCache(cacheLifeTime, self.__updateRssCache)

    def __updateRssCache(self):
        """Method used to update the rssCache.

        It will try 5 times to contact the RSS before giving up
        """

        meta = {"columns": ["Name", "Status", "VO"]}

        for ti in range(5):
            rawCache = self.rsClient.selectStatusElement("Site", "Status", meta=meta)
            if rawCache["OK"]:
                break
            self.log.warn("Can't get resource's status", rawCache["Message"] + "; trial %d" % ti)
            sleep(math.pow(ti, 2))
            self.rsClient = ResourceStatusClient()

        if not rawCache["OK"]:
            return rawCache
        return S_OK(getCacheDictFromRawData(rawCache["Value"]))

    def getSiteStatuses(self, siteNames=None):
        """
        Method that queries the database for status of the sites in a given list.
        A single string site name may also be provides as "siteNames"
        If the input is None, it is interpreted as * ( all ).

        If match is positive, the output looks like::

          {
           'test1.test1.org': 'Active',
           'test2.test2.org': 'Banned',
          }

        Examples::

          >>> siteStatus.getSiteStatuses( ['test1.test1.uk', 'test2.test2.net', 'test3.test3.org'] )
              S_OK( { 'test1.test1.org': 'Active', 'test2.test2.net': 'Banned', 'test3.test3.org': 'Active' }  )
          >>> siteStatus.getSiteStatuses( 'NotExists')
              S_ERROR( ... ))
          >>> siteStatus.getSiteStatuses( None )
              S_OK( { 'test1.test1.org': 'Active',
                      'test2.test2.net': 'Banned', },
                      ...
                    }
                  )

        :param siteNames: name(s) of the sites to be matched
        :type siteNames: list, str
        :return: S_OK() || S_ERROR()
        """

        if self.rssFlag:
            return self.__getRSSSiteStatus(siteNames)
        else:
            siteStatusDict = {}
            wmsAdmin = WMSAdministratorClient()
            if siteNames:
                if isinstance(siteNames, six.string_types):
                    siteNames = [siteNames]
                for siteName in siteNames:
                    result = wmsAdmin.getSiteMaskStatus(siteName)
                    if not result["OK"]:
                        return result
                    else:
                        siteStatusDict[siteName] = result["Value"]
            else:
                result = wmsAdmin.getSiteMaskStatus()
                if not result["OK"]:
                    return result
                else:
                    siteStatusDict = result["Value"]

            return S_OK(siteStatusDict)

    def __getRSSSiteStatus(self, siteName=None):
        """Gets from the cache or the RSS the Sites status. The cache is a
            copy of the DB table. If it is not on the cache, most likely is not going
            to be on the DB.

            There is one exception: item just added to the CS, e.g. new Element.
            The period between it is added to the DB and the changes are propagated
            to the cache will be inconsistent, but not dangerous. Just wait <cacheLifeTime>
            minutes.

        :param siteName: name of the site
        :type siteName: str

        :return: dict
        """

        cacheMatch = self.rssCache.match(siteName, "", "", "all")  # sites have VO="all".

        self.log.debug("__getRSSSiteStatus")
        self.log.debug(cacheMatch)

        return cacheMatch

    def getUsableSites(self, siteNames=None):
        """
        Returns all sites that are usable if their
        statusType is either Active or Degraded; in a list.

        examples
          >>> siteStatus.getUsableSites( ['test1.test1.uk', 'test2.test2.net', 'test3.test3.org'] )
              S_OK( ['test1.test1.uk', 'test3.test3.org'] )
          >>> siteStatus.getUsableSites( None )
              S_OK( ['test1.test1.uk', 'test3.test3.org', 'test4.test4.org', 'test5.test5.org', ...] )
          >>> siteStatus.getUsableSites( 'NotExists' )
              S_ERROR( ... )

        :Parameters:
          **siteNames** - `List` or `str`
            name(s) of the sites to be matched

        :return: S_OK() || S_ERROR()
        """

        siteStatusDictRes = self.getSiteStatuses(siteNames)
        if not siteStatusDictRes["OK"]:
            return siteStatusDictRes
        siteStatusList = [x[0] for x in siteStatusDictRes["Value"].items() if x[1] in ["Active", "Degraded"]]

        return S_OK(siteStatusList)

    def getSites(self, siteState="Active"):
        """
        By default, it gets the currently active site list

        examples
          >>> siteStatus.getSites()
              S_OK( ['test1.test1.uk', 'test3.test3.org'] )
          >>> siteStatus.getSites( 'Active' )
              S_OK( ['test1.test1.uk', 'test3.test3.org'] )
          >>> siteStatus.getSites( 'Banned' )
              S_OK( ['test0.test0.uk', ... ] )
          >>> siteStatus.getSites( 'All' )
              S_OK( ['test1.test1.uk', 'test3.test3.org', 'test4.test4.org', 'test5.test5.org'...] )
          >>> siteStatus.getSites( None )
              S_ERROR( ... )

        :Parameters:
          **siteState** - `String`
            state of the sites to be matched

        :return: S_OK() || S_ERROR()
        """

        if not siteState:
            return S_ERROR(DErrno.ERESUNK, "siteState parameter is empty")

        siteStatusDictRes = self.getSiteStatuses()
        if not siteStatusDictRes["OK"]:
            return siteStatusDictRes

        if siteState.capitalize() == "All":
            # if no siteState is set return everything
            siteList = list(siteStatusDictRes["Value"])

        else:
            # fix case sensitive string
            siteState = siteState.capitalize()
            allowedStateList = ["Active", "Banned", "Degraded", "Probing", "Error", "Unknown"]
            if siteState not in allowedStateList:
                return S_ERROR(errno.EINVAL, "Not a valid status, parameter rejected")

            siteList = [x[0] for x in siteStatusDictRes["Value"].items() if x[1] == siteState]

        return S_OK(siteList)

    def setSiteStatus(self, site, status, comment="No comment"):
        """
        Set the status of a site in the 'SiteStatus' table of RSS

        examples
          >>> siteStatus.banSite( 'site1.test.test' )
              S_OK()
          >>> siteStatus.banSite( None )
              S_ERROR( ... )

        :Parameters:
          **site** - `String`
            the site that is going to be banned
          **comment** - `String`
            reason for banning

        :return: S_OK() || S_ERROR()
        """

        if not status:
            return S_ERROR(DErrno.ERESUNK, "status parameter is empty")

        # fix case sensitive string
        status = status.capitalize()
        allowedStateList = ["Active", "Banned", "Degraded", "Probing", "Error", "Unknown"]

        if status not in allowedStateList:
            return S_ERROR(errno.EINVAL, "Not a valid status, parameter rejected")

        if self.rssFlag:
            result = getProxyInfo()
            if result["OK"]:
                tokenOwner = result["Value"]["username"]
            else:
                return S_ERROR("Unable to get user proxy info %s " % result["Message"])

            tokenExpiration = datetime.utcnow() + timedelta(days=1)

            self.rssCache.acquireLock()
            try:
                result = self.rsClient.modifyStatusElement(
                    "Site",
                    "Status",
                    status=status,
                    name=site,
                    tokenExpiration=tokenExpiration,
                    reason=comment,
                    tokenOwner=tokenOwner,
                )
                if result["OK"]:
                    self.rssCache.refreshCache()
                else:
                    _msg = "Error updating status of site %s to %s" % (site, status)
                    gLogger.warn("RSS: %s" % _msg)

            # Release lock, no matter what.
            finally:
                self.rssCache.releaseLock()

        else:
            if status in ["Active", "Degraded"]:
                result = WMSAdministratorClient().allowSite()
            else:
                result = WMSAdministratorClient().banSite()

        return result


def getCacheDictFromRawData(rawList):
    """
    Formats the raw data list, which we know it must have tuples of four elements.
    ( element1, element2 ) into a dictionary of tuples with the format
    { ( element1 ): element2 )}.
    The resulting dictionary will be the new Cache.

    It happens that element1 is elementName,
                    element4 is status.

    :Parameters:
      **rawList** - `list`
        list of three element tuples [( element1, element2 ),... ]

    :return: dict of the form { ( elementName ) : status, ... }
    """

    res = {}
    for entry in rawList:
        res.update({(entry[0]): entry[1]})

    return res
